@@ -1,7 +1,7 @@
 # Accelerator for pip, the Python package manager.
 #
 # Author: Peter Odding <peter.odding@paylogic.eu>
-# Last Change: June 15, 2013
+# Last Change: June 16, 2013
 # URL: https://github.com/paylogic/pip-accel
 #
 # TODO Permanently store logs in the pip-accel directory (think about log rotation).
@@ -16,47 +16,44 @@ taking a look at the following functions:
 
 - :py:func:`unpack_source_dists`
 - :py:func:`download_source_dists`
-- :py:func:`build_binary_dists`
+- :py:func:`build_missing_binary_dists`
 - :py:func:`install_requirements`
 """
 
 # Semi-standard module versioning.
-__version__ = '0.8.16'
+__version__ = '0.9.2'
 
 # Standard library modules.
-import logging
 import os
 import os.path
 import pkg_resources
 import pwd
 import re
+import shutil
+import subprocess
 import sys
 import tarfile
+import tempfile
 import textwrap
 import time
 import urllib
 import urlparse
 
+# Internal modules.
+from pip_accel.deps import sanity_check_dependencies
+from pip_accel.logger import logger
+
 # External dependencies.
-from coloredlogs import ColoredStreamHandler
 from pip.backwardcompat import string_types
 from pip.baseparser import create_main_parser
 from pip.commands.install import InstallCommand
 from pip.exceptions import DistributionNotFound, InstallationError
+from pip.log import logger as pip_logger
 from pip.status_codes import SUCCESS
 
-# Initialize the logging subsystem.
-logger = logging.getLogger('pip-accel')
-logger.setLevel(logging.INFO)
-logger.addHandler(ColoredStreamHandler())
-
 # Whether a user is directly looking at the output.
 INTERACTIVE = os.isatty(1)
 
-# Check if the operator requested verbose output.
-if '-v' in sys.argv or 'PIP_ACCEL_VERBOSE' in os.environ:
-    logger.setLevel(logging.DEBUG)
-
 # Find the environment where requirements are to be installed.
 ENVIRONMENT = os.path.abspath(os.environ.get('VIRTUAL_ENV', sys.prefix))
 
@@ -133,23 +130,24 @@ def main():
             try:
                 requirements = unpack_source_dists(arguments)
             except DistributionNotFound:
-                logger.warn("Warning: We don't have all source distributions yet!")
+                logger.warn("We don't have all source distributions yet!")
                 download_source_dists(arguments)
             else:
                 if not requirements:
                     logger.info("No unsatisfied requirements found, probably there's nothing to do.")
                 else:
-                    if build_binary_dists(requirements) and install_requirements(requirements):
+                    if build_missing_binary_dists(requirements) and install_requirements(requirements):
                         logger.info("Done! Took %s to install %i package%s.", main_timer, len(requirements), '' if len(requirements) == 1 else 's')
                     else:
                         sys.exit(1)
                 return
+            logger.warn("pip failed, retrying (%i/%i) ..", i + 1, MAX_RETRIES)
     except InstallationError:
         # Abort early when pip reports installation errors.
-        logger.fatal("Error: Pip reported unrecoverable installation errors. Please fix and rerun!")
+        logger.fatal("pip reported unrecoverable installation errors. Please fix and rerun!")
         sys.exit(1)
     # Abort when after N retries we still failed to download source distributions.
-    logger.fatal("Error: External command failed %i times, aborting!" % MAX_RETRIES)
+    logger.fatal("External command failed %i times, aborting!" % MAX_RETRIES)
     sys.exit(1)
 
 def print_usage():
@@ -249,9 +247,9 @@ def download_source_dists(arguments):
         run_pip(arguments + ['--no-install'], use_remote_index=True)
         logger.info("Finished downloading source distributions in %s.", download_timer)
     except Exception, e:
-        logger.warn("Error: pip raised an exception while downloading source distributions: %s.", e)
+        logger.warn("pip raised an exception while downloading source distributions: %s.", e)
 
-def find_binary_dists():
+def find_cached_binary_dists():
     """
     Find all previously cached binary distributions.
 
@@ -276,7 +274,7 @@ def find_binary_dists():
         logger.debug(" - %s (%s, %s) in %s.", name, version, pyversion, filename)
     return distributions
 
-def build_binary_dists(requirements):
+def build_missing_binary_dists(requirements):
     """
     Convert source distributions to binary distributions.
 
@@ -287,7 +285,7 @@ def build_binary_dists(requirements):
               ``False`` otherwise (probably because of missing binary
               dependencies like system libraries).
     """
-    existing_binary_dists = find_binary_dists()
+    existing_binary_dists = find_cached_binary_dists()
     logger.info("Building binary distributions ..")
     pyversion = get_python_version()
     for name, version, directory in requirements:
@@ -299,30 +297,73 @@ def build_binary_dists(requirements):
         # Make sure the source distribution contains a setup script.
         setup_script = os.path.join(directory, 'setup.py')
         if not os.path.isfile(setup_script):
-            logger.warn("Warning: Package %s (%s) is not a source distribution.", name, version)
+            logger.warn("Package %s (%s) is not a source distribution.", name, version)
             continue
-        old_directory = os.getcwd()
-        # Build a binary distribution.
-        os.chdir(directory)
-        logger.info("Building binary distribution of %s (%s) ..", name, version)
-        status = (os.system('"%s/bin/python" setup.py bdist_dumb --format=gztar' % ENVIRONMENT) == 0)
-        os.chdir(old_directory)
-        if not status:
-            logger.error("Failed to build binary distribution of %s! (%s)!", name, version)
-            return False
-        # Move the generated distribution to the binary index.
-        filenames = os.listdir(os.path.join(directory, 'dist'))
-        if len(filenames) != 1:
-            logger.error("Error: Build process did not result in one binary distribution! (matches: %s)", filenames)
-            return False
-        cache_file = '%s:%s:%s.tar.gz' % (name, version, pyversion)
-        logger.info("Copying binary distribution %s to cache as %s.", filenames[0], cache_file)
-        cache_binary_distribution(os.path.join(directory, 'dist', filenames[0]),
-                                  os.path.join(binary_index, cache_file))
+        # Try to build the binary distribution.
+        if not build_binary_dist(name, version, directory, pyversion):
+            if not (sanity_check_dependencies(name) and build_binary_dist(name, version, directory, pyversion)):
+                return False
     logger.info("Finished building binary distributions.")
     return True
 
-def cache_binary_distribution(input_path, output_path):
+def build_binary_dist(name, version, directory, pyversion):
+    """
+    Convert a single, unpacked source distribution to a binary distribution.
+
+    :param name: The name of the requirement to build.
+    :param version: The version of the requirement to build.
+    :param directory: The directory where the unpacked sources of the
+                      requirement are available.
+
+    :returns: ``True`` if we succeed in building a binary distribution,
+              ``False`` otherwise (probably because of missing binary
+              dependencies like system libraries).
+    """
+    # Cleanup previously generated distributions.
+    dist_directory = os.path.join(directory, 'dist')
+    if os.path.isdir(dist_directory):
+        logger.info("Cleaning up previously generated distributions in %s ..", dist_directory)
+        shutil.rmtree(dist_directory)
+    # Let the user know what's going on.
+    build_text = "Building binary distribution of %s (%s) .." % (name, version)
+    logger.info("%s", build_text)
+    # Compose the command line needed to build the binary distribution.
+    command_line = '"%s/bin/python" setup.py bdist_dumb --format=gztar' % ENVIRONMENT
+    logger.debug("Executing external command: %s", command_line)
+    # Redirect all output of the build to a temporary file.
+    fd, temporary_file = tempfile.mkstemp()
+    command_line = '%s > "%s" 2>&1' % (command_line, temporary_file)
+    # Start the build.
+    build = subprocess.Popen(['sh', '-c', command_line], cwd=directory)
+    # Wait for the build to finish.
+    if INTERACTIVE:
+        # Provide feedback to the user in the mean time.
+        spinner = Spinner(build_text)
+        while build.poll() is None:
+            spinner.step()
+            time.sleep(0.1)
+        spinner.clear()
+    else:
+        build.wait()
+    # Make sure the build succeeded.
+    if build.returncode != 0:
+        logger.error("Failed to build binary distribution of %s! (version: %s)", name, version)
+        with open(temporary_file) as handle:
+            logger.info("Build output (will probably provide a hint as to what went wrong):\n%s", handle.read())
+        return False
+    # Move the generated distribution to the binary index.
+    filenames = os.listdir(dist_directory)
+    if len(filenames) != 1:
+        logger.error("Build process did not result in one binary distribution! (matches: %s)", filenames)
+        return False
+    cache_file = '%s:%s:%s.tar.gz' % (name, version, pyversion)
+    logger.info("Copying binary distribution %s to cache as %s.", filenames[0], cache_file)
+    cache_binary_dist(os.path.join(directory, 'dist', filenames[0]),
+                      os.path.join(binary_index, cache_file))
+    logger.info("Finished building binary distributions.")
+    return True
+
+def cache_binary_dist(input_path, output_path):
     """
     Transform a binary distribution archive created with ``python setup.py
     bdist_dumb --format=gztar`` into a form that can be cached for future use.
@@ -349,7 +390,7 @@ def cache_binary_distribution(input_path, output_path):
         elif not member.isdir():
             modified_pathname = os.path.relpath(absolute_pathname, ENVIRONMENT)
             if os.path.isabs(modified_pathname):
-                logger.warn("Warning: Failed to transform pathname in binary distribution to relative path! (original: %r, modified: %r)",
+                logger.warn("Failed to transform pathname in binary distribution to relative path! (original: %r, modified: %r)",
                             original_pathname, modified_pathname)
             else:
                 logger.debug("Transformed %r -> %r.", original_pathname, modified_pathname)
@@ -376,24 +417,25 @@ def install_requirements(requirements, install_prefix=ENVIRONMENT):
               binary distribution archives, ``False`` otherwise.
     """
     install_timer = Timer()
-    existing_binary_dists = find_binary_dists()
+    existing_binary_dists = find_cached_binary_dists()
     pyversion = get_python_version()
     logger.info("Installing from binary distributions ..")
     for name, version, directory in requirements:
         filename = existing_binary_dists.get((name.lower(), version, pyversion))
         if not filename:
-            logger.error("Error: No binary distribution of %s (%s) available!", name, version)
+            logger.error("No binary distribution of %s (%s) available!", name, version)
             return False
-        install_binary_dist(filename, install_prefix=install_prefix)
+        install_binary_dist(name, filename, install_prefix=install_prefix)
     logger.info("Finished installing all requirements in %s.", install_timer)
     return True
 
-def install_binary_dist(filename, install_prefix=ENVIRONMENT):
+def install_binary_dist(package, filename, install_prefix=ENVIRONMENT):
     """
     Install a binary distribution created with ``python setup.py bdist`` into
     the given prefix (a directory like ``/usr``, ``/usr/local`` or a virtual
     environment).
 
+    :param package: The name of the package to install.
     :param filename: The pathname of the tar archive.
     :param install_prefix: The "prefix" under which the requirements should be
                            installed. This will be a pathname like ``/usr``,
@@ -408,7 +450,10 @@ def install_binary_dist(filename, install_prefix=ENVIRONMENT):
     #     places based on the "seed" environment.
     install_timer = Timer()
     python = os.path.join(install_prefix, 'bin', 'python')
-    logger.info("Installing binary distribution %s to %s ..", filename, install_prefix)
+    pip = os.path.join(install_prefix, 'bin', 'pip')
+    if os.system('"%s" uninstall --yes "%s" >/dev/null 2>&1' % (pip, package)) == 0:
+        logger.info("Uninstalled previously installed package %s.", package)
+    logger.info("Installing package %s from binary distribution %s to %s ..", package, filename, install_prefix)
     archive = tarfile.open(filename, 'r:gz')
     for member in archive.getmembers():
         install_path = os.path.join(install_prefix, member.name)
@@ -481,7 +526,7 @@ def run_pip(arguments, use_remote_index):
             break
     else:
         command_line = ['pip'] + arguments
-    logger.info("Executing command: %s.", ' '.join(command_line))
+    logger.info("Executing command: %s", ' '.join(command_line))
     parser = create_main_parser()
     pip = CustomInstallCommand(parser)
     initial_options, args = parser.parse_args(command_line[1:])
@@ -497,19 +542,32 @@ def run_pip(arguments, use_remote_index):
 class CustomInstallCommand(InstallCommand):
 
     """
-    Required to call ``pip`` as a Python module instead of a subprocess.
-
-    The ``pip.commands.install.InstallCommand.run()`` method returns a
-    ``RequirementSet`` object which ``pip-accel`` is interested in, however
-    ``pip.basecommand.Command.main()`` swallows the requirement set (based on
-    my reading of the pip 1.3.x source code).
-
-    To work around the problem described above, we subclass ``InstallCommand``
-    to wrap the run() method. Yes this is a bit sneaky, but I don't fancy
-    reimplementing ``pip.basecommand.Command.main()`` inside ``pip-accel``!
+    Subclass of ``pip.commands.install.InstallCommand`` that makes it easier to
+    run ``pip install`` from Python.
     """
 
+    def main(self, *args, **kw):
+        """
+        ``pip.basecommand.Command.main()`` expects to be executed only once; it
+        unconditionally executes ``pip.log.logger.consumers.extend()``. This
+        means that when we run ``pip`` more than once we'll cause it to repeat
+        its output as many times as we executed a ``pip install`` command. We
+        wrap ``main()`` to explicitly reset the list of consumers.
+        """
+        pip_logger.consumers = []
+        return super(CustomInstallCommand, self).main(*args, **kw)
+
+
     def run(self, *args, **kw):
+        """
+        The method ``pip.commands.install.InstallCommand.run()`` returns a
+        ``RequirementSet`` object which ``pip-accel`` is interested in, however
+        ``pip.basecommand.Command.main()`` (the caller of ``run()``) swallows
+        the requirement set (based on my reading of the pip 1.3.x source code).
+        We wrap ``run()`` so that we can intercept the requirement set. This is
+        a bit sneaky, but I don't fancy reimplementing large parts of
+        ``pip.basecommand.Command.main()`` inside ``pip-accel``!
+        """
         original_method = super(CustomInstallCommand, self).run
         try:
             self.intercepted_exception = None
@@ -635,5 +693,36 @@ class Timer:
         """
         return "%.2f seconds" % self.elapsed_time
 
+class Spinner:
+
+    """
+    Show a "spinner" on the terminal to let the user know that we're busy
+    building a package's binary distribution without dumping all of the
+    output on the terminal.
+    """
+
+    def __init__(self, label):
+        self.label = label
+        self.states = ['-', '\\', '|', '/']
+        self.counter = 0
+
+    def step(self):
+        """
+        Advance the spinner by one step without starting a new line, causing
+        an animated effect which is very simple but much nicer than waiting
+        for a prompt which is completely silent for a long time.
+        """
+        state = self.states[self.counter % len(self.states)]
+        sys.stderr.write("\r %s %s " % (state, self.label))
+        self.counter += 1
+
+    def clear(self):
+        """
+        Clear the spinner. The next line which is shown on the standard
+        output or error stream after calling this method will overwrite the
+        line that used to show the spinner.
+        """
+        sys.stderr.write("\r")
+
 if __name__ == '__main__':
     main()
